package Ightaria;

import java.awt.*;
import java.applet.*;
import java.awt.event.*;
import java.lang.Math;

import Globals.*;
import DrawTexts.*;
import CalculusAndVectors.*;
import Statistics.Stat;




/** Class "transitional" between Vessel and other
    classes
*/
public class Bridge {



//SOME PAR:
//do2 clumsy for derivator:
//This two flags must cause their actions at once and
//this actions must follow by derivator if it is set:
boolean flagUpdateCurrentVesselPosAOrientation = false; //Also for resetting speed.
boolean flagUpdateScenePopulation = false;

boolean flagUpdatePosAScene       = false; //Also for resetting speed.

static public int frameNumber;             //Replica of Vessel.frameNumber.


/** Modes */
boolean autoPilot;




//=====================================================================
// L o c a l   N a v i g a t i o n
//---------------------------------------------------------------------
/** Elevation */  double x3;  
                  double elevationStep;
/** For stepXY */ double defaultStep =  5.;
                  static double stepXY;
/** For stepXY */ static double stepLimit;

static double ceiling;
double floor             =  20.;
double angleStepLowLimit =  1.e-15;
double AngleStepDefault  =  Math.PI/36.; 
double AngleStep;

/** Increment. For Course and AutoPilot. */
public static double  vFpAccIncr=0;            
public static double  oAccomulatedPath;
boolean MovedAndStoppedIsInProgress;

/** Driver's turn in altitude direction.
    Only for display. Irrelevant for
    path and motions.
*/
double  dAltiIncr;            
/** Driver's head rotation over axis x is enabled if true */
static boolean engAltitude;

public static String navigationMessage;

//---------------------------------------------------------------------
// L o c a l   N a v i g a t i o n
//=====================================================================






//=======================================
//Driver's Panels
//- - - - - - - - - - - - - - - - - - - -
static final int spSys     = SPad.a_;       //Computer's Health.
static final int spModes   = SPad.b_;       //Driver's Level. Dialog Modes. 
static final int spFDigits = SPad.W_;    
static final int spMDigits = SPad.E_;
static final int spFPad    = SPad.c_;
static final int spMPad    = SPad.d_;
static final int spNav     = SPad.S_;       //Navigation
static final int spDialog  = SPad.N_;       //Driver's Dialog
static final int spMain    = SPad.C_;
//=======================================

//Interface to Observer
boolean displayAxes = false;
/** Triggers painting of four SPads:
    W_, E_, c_, d_
*/



/////////////////////////////
// Bridge Events:
//===========================
public boolean keyEventHappened;
public static final int KEY_UP_DOWN_NONE      = 0;  //KeyEvent.VK_LEFT;
public static final int KEY_UP_DOWN_DIGESTED  = 1;  //KeyEvent.VK_RIGHT;
public static final int KEY_UP_DOWN_COLLISION = 2;  //KeyEvent.VK_PAGE_DOWN;
public int keyUpDownCode;

int  keyID;        //(int) getID. Method inherited from class java.awt.AWTEvent
int  keyCode;      //(int) getKeyCode
int  keyModifiers; //(int) getModifiersEx

int  eventToScenario; 
int  flagScrollText;
/////////////////////////////



//B r i d g e   S c e n a r i o   e l e m e n t s
static final int exitNOT_ORDERED = 0;
static final int exitORDERED     = 1;
static final int exitCOMPLETED   = 2;
static       int exitFlag        = exitNOT_ORDERED;

static       int exitDURATION    = 10;
static       int exitLastFrame;





//===============================================
// Resetters
//-----------------------------------------------
public static void setLevelLimits() {
       switch(Scenario.level) {
       case Scenario.levAncestor:
       case Scenario.levJustBorn:       
       case Scenario.levNavigationExpert:
            stepLimit   = 20.;
            ceiling     = 500.;
            break;
       case Scenario.levHFConqueror:
            //Don't be greedy:
            stepLimit   = 2.*Math.PI*HFold.hfRadius;
            ceiling = 10000.;
            break;
       }
       if(Scenario.cheatEnabled){
          stepLimit   = 20.*Math.PI*HFold.hfRadius;
          ceiling = 1000000.;
       }else{
          if( stepXY > stepLimit + 1. ) {
              stepXY = stepLimit;
              //do2 cannot do this because collision:
              //x3 < ceiling ...
          }
       }
}


public void restart() {

   GS.con("Restarting Bridge ... ");
   Vessel.frameNumber = 1;
   setLevelLimits();

   //do1 Must test for collision immediately:
   x3 =120;  

   vFpAccIncr=0;
   oAccomulatedPath =  0;
   MovedAndStoppedIsInProgress = false;
   keyEventHappened            = false;
   keyUpDownCode               = KEY_UP_DOWN_NONE;
   eventToScenario  = Scenario.EV_NOEVENT; 
   flagScrollText   = 0;
   Scenario.setText(Scenario.TXT_NOTHING);

   HFold.restart();
   restore_defaults();
   Pop.cancelAllMeasurements();

}

/** When updating this method, don't forget to 
    put what it does into Help.txt */
public void restore_defaults() {

   autoPilot      = false;
   engAltitude    = false;

   elevationStep  = 10.;
   stepXY         = defaultStep;  //Horizontal Speed
   AngleStep      = AngleStepDefault;
   dAltiIncr      = 0.;
   
   flagUpdatePosAScene = true;
   HFold.reset_defaults();

}

//-----------------------------------------------
// Resetters
//===============================================





//IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII
//	C O N S T R U C T O R
//========================================
public Bridge() {
       //SPad.initializePads();
       HFold.initiate();
       Pop.initiate();  
       Scenario.initiate(); 
       restart();
       flagUpdatePosAScene = true;
} 
//IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII



//================================================================
//  S u p p l y   a n d   r e l e a s e   r e s o u r c e s
//----------------------------------------------------------------
public void supplyGraphics() {
       HFold.supplyGraphics();
       Pop.supplyGraphics();
}
public void releaseGraphics() {
       HFold.releaseGraphics();
       Pop.releaseGraphics();
       SPad.releaseGraphicsFromCollection();
}
//================================================================





//RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR
// 		RUN
//============================================================================================
public void runFrameStep(
                int  keyID,
                int  keyCode,     
                int  keyModifiers, 
                int  frameNumber
) {

      this.frameNumber = frameNumber;

      this.keyID            = keyID ;
      this.keyCode          = keyCode;
      this.keyModifiers     = keyModifiers;

      Scenario.indicateLevel();
        
      if( keyEventHappened ) { 
          //GS.con("Bridge. KeyEvent Happened. keyCode=" + keyCode);
          Stat.keyEvents++;
          keyEventHappened = false;
          if( exitORDERED != Bridge.exitFlag ) keyPressedHandler();
      }


      if( flagUpdatePosAScene ){
          flagUpdatePosAScene = false;
          updatePosAScene();
      }else{ 
          if( flagUpdateCurrentVesselPosAOrientation ) {
              HFold.updatePathLastMotion();
          } 
          if( flagUpdateScenePopulation) {
              Pop.PopulateScene(); 
          }
          if( flagUpdateCurrentVesselPosAOrientation ||
              flagUpdateScenePopulation ) {
              HFold.runAngleminerAnDerivator();
          } 
          flagUpdateCurrentVesselPosAOrientation = false;
          flagUpdateScenePopulation = false;
     }


     if( MovedAndStoppedIsInProgress ) {
         //GS.con("Bridge. Moved and stopped ... ");
         HFold.FinalizeStop();
         HFold.runAngleminerAnDerivator();
         keyUpDownCode = KEY_UP_DOWN_NONE;
         MovedAndStoppedIsInProgress = false;
     } 




      //Navigational part finished. Do dialog, messages, paint.

         
      if( navigationMessage != "" ) {
          mesnavF(navigationMessage);
          navigationMessage = "";
      }

      Scenario.dialog("Put Bridge Response Here", eventToScenario);
      eventToScenario = Scenario.EV_NOEVENT;

      HFold.paintMotionsAndMeasure(frameNumber);

      //This is a bad programming: But we need for this
      //to put a Welcome message:
      SPad.sPad[SPad.C_].wrc(" ");
      if( 1 == frameNumber ) {
          Scenario.setupWelcomeMessage(frameNumber);
      }
      Scenario.printGrandMessage(frameNumber,0);  //Print Far Message.

      //Third true fills background:
      Pop.grHill.paintHill ( 
          (int)x3, frameNumber, false, false, 
          displayAxes, dAltiIncr
      );            

      if( GS.deb ) HFold.displayVesselBasis();

      Scenario.printGrandMessage(frameNumber,1); //Print Front Message
            
      //---------------------------------------------
      //Output texts and sketches on top of pads ...
      //- - - - - - - - - - - - - - - - - - - - - - -
      Scenario.displayText(flagScrollText);
      flagScrollText = 0; //Drop flag.
      //---------------------------------------------

      SPad.paintTalk();  

}//RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR








private void keyPressedHandler() { 

      int i;
      double st, safeSmallStep;
      double x,y,z,w;        
      boolean encourageObserver = true;
      //Give compiler something to eat:
      Hill h = Pop.hills[1]; 
      //Specials:
      double wHVp;
      double wOVp;


      //GS.con(">>Bridge. KeyHandler:  " + keyCode + "." + keyID + " keyModifiers=" + keyModifiers);
      //GS.con(">>Bridge. VK_LESS=" + KeyEvent.VK_LESS + " CTRL_DOWN_MASK=" + KeyEvent.CTRL_DOWN_MASK);


      switch (keyModifiers) {
      case KeyEvent.CTRL_DOWN_MASK:
           switch (keyCode) {

                     //These are goodies: let them stay here.
                     //VK_F1 VK_EXCLAMATION_MARK 
                     //VK_SPACE:
                     //VK_HOME VK_INSERT
                     //VK_LEFT_PARENTHESIS ( VK_RIGHT_PARENTHESIS
                     //VK_COLON
                     //VK_OPEN_BRACKET
                     //case KeyEvent.VK_EQUALS:
                     //     GS.dcon("Key =? ");
                     //     break;
  
           //R e s t o r e 
           case KeyEvent.VK_R: 
                restart();
                navigationMessage="SPACE RENTERED";
                break;

           case KeyEvent.VK_0: 
                restore_defaults(); 
                navigationMessage="DEFAULTS RESTORED";
                break;

           case KeyEvent.VK_A: 
                SPad.sPad[SPad.N_].wrc(Scenario.COPYRIGHT[0]);
                SPad.sPad[SPad.N_].wr (Scenario.COPYRIGHT[1]);
                //Protect email from spam:
                SPad.sPad[SPad.N_].wr (Scenario.COPYRIGHT[2].replace('-','@'));
                restore_defaults(); 
                break;





           //S c r o l l   t e x t   U p / D o w n
           case KeyEvent.VK_OPEN_BRACKET:
                flagScrollText = -1; //Down
                return;
           case KeyEvent.VK_CLOSE_BRACKET:
                flagScrollText = 1;  //Up
                return;
           case KeyEvent.VK_EQUALS:
                //do1b More precise method must be:
                //i = KeyEvent.SHIFT_DOWN_MASK | KeyEvent.CTRL_DOWN_MASK;
                //if( keyModifiers & i == i ) {
                //Or must check keyChar
                //because theoretically "+" can be not
                //on "=" key on some keyboards.
                flagScrollText = 2; //Page Down
                return;
           case KeyEvent.VK_MINUS:
                flagScrollText = -2;  //Page Up
                return;



           //-------------------------------------------
           //S e t   m o t i o n s   a c c u r a c y
           //- - - - - - - - - - - - - - - - - - - - - -
           case KeyEvent.VK_F:
                w = stepXY * 2;
                if( w < stepLimit + 1. ) {
                    stepXY = w;
                    navigationMessage ="SPEED UP";
                }
                break;
           case KeyEvent.VK_S:
                w = stepXY / 2;
                if( w > 1. ) {
                    stepXY = w;
                    navigationMessage="SLOW DOWN";
                }
                break;

           case KeyEvent.VK_C: //Crude
                w = AngleStep * 10.;
                if( w <= Vec.DPI ) {
                    AngleStep = w;
                    navigationMessage="CRUDE AZI";
                }else{
                    navigationMessage="AZI STEP RESET TO DEFAULT";
                    AngleStep = AngleStepDefault;
                }    
                break;
           case KeyEvent.VK_N: //"Nicer"
                w = AngleStep / 10.;
                if( w > angleStepLowLimit ) {
                    AngleStep  = w;
                    navigationMessage="NICE AZI";
                }else{
                    navigationMessage="AZI STEP RESET TO DEFAULT";
                    AngleStep = AngleStepDefault;
                }    
                break;
           //- - - - - - - - - - - - - - - - - - - - - -
           //S e t   m o t i o n s   a c c u r a c y
           //-------------------------------------------




           //------------------------------------------
           // Special observation switches.
           //- - - - - - - - - - - - - - - - - - - - - 

           case KeyEvent.VK_M:
                Pop.toggleMeasurementedHill();
                break;            


           case KeyEvent.VK_L:   
                //Toggle measured Hill ...
                Pop.toggleMeasurementedHill();
                //... and Turn driver to hill.
           case KeyEvent.VK_D:
                i = Pop.iMeasuredHill;
                if( i > 0 ) {
                    h = Pop.hills[Pop.iMeasuredHill];
                    wHVp = h.twin_hVp(false);
                    wOVp = Vec.fixAzi(HFold.oFp - HFold.vFp);
                    //Update of driver is instant:
                    HFold.oFp = Vec.fixAzi(HFold.oFp - wOVp + wHVp); 
                    if(!autoPilot) {
                       //Update of vessell goes trough normal rotation:
                       vFpAccIncr = Vec.fixAzi(vFpAccIncr + wHVp);
                    }
                }
                navigationMessage="DRIRVER TURNED TO " + h.measuredName();
                flagUpdatePosAScene = true;
                HFold.doPaintScratchMaps = true;
                break;


           case KeyEvent.VK_T:
                i = Pop.iMeasuredHill;
                if( i == 0 ) {
                    navigationMessage="HILL NOT SELECTED. NO REPOSITION.";
                    return;
                }
                h = Pop.hills[Pop.iMeasuredHill];
                if( !h.inSceneRange && 
                    Scenario.levHFConqueror > Scenario.level && 
                    !Scenario.cheatEnabled ) {
                    navigationMessage="STAR TOO FAR. NO REPOSITION.";
                    return;
                }
                if( HFold.sphereNavigationTrustedGeneralAccuracy/2. > 
                    h.oDistance ) {
                    navigationMessage="ENOUGH CLOSE. NO REPOS.";
                    return;
                }
                wHVp = h.twin_hVp(false);
                vFpAccIncr = Vec.fixAzi(vFpAccIncr + wHVp);
                if( !autoPilot) {
                    //Update of driver is instant based on coming V. update:
                    HFold.oFp = Vec.fixAzi(HFold.vFpS + vFpAccIncr); 
                }
                //true simulates move:
                if( doTriesToAvoidCollisions(h.oDistance, true) ) {
                    navigationMessage="HILL REPOSITIONED ON " + h.measuredName();
                }else{
                    navigationMessage="NO REPOSITION. HILL TOO HIGH: " + h.measuredName();
                }

                //This is a vital to set:
                MovedAndStoppedIsInProgress = true;

                HFold.doPaintScratchMaps = true;
                break;



           case KeyEvent.VK_V:   //Turn Vessel to hill.
                                 //No measur toggling.
                i = Pop.iMeasuredHill;
                if( i == 0 ) {
                    navigationMessage="NO REDIRECTION. HILL NOT SELECTED";
                    return;
                }
                h = Pop.hills[i];
                wHVp = h.twin_hVp(false);
                vFpAccIncr = Vec.fixAzi(vFpAccIncr + wHVp);
                if(!autoPilot) {
                   //Update of driver is instant based on coming V. update:
                   HFold.oFp = Vec.fixAzi(HFold.vFpS + vFpAccIncr); 
                }
                navigationMessage="DIRECTED TO " + h.measuredName();
                flagUpdatePosAScene = true;
                HFold.doPaintScratchMaps = true;
                break;





          case KeyEvent.VK_P:
               autoPilot = !autoPilot;
               navigationMessage="PILOT SWITCH";
               break;   

          case KeyEvent.VK_O:   //Synchronize vessel to driver.
               navigationMessage="TURN VESSEL TO DRIVER";
               //GS.con("Sync V to D: oFp=" + HFold.oFp + " vFp=" + HFold.vFp);
               vFpAccIncr = Vec.fixAzi(vFpAccIncr + HFold.oFp - HFold.vFp);
               flagUpdatePosAScene = true;
               HFold.doPaintScratchMaps = true;
               break;

          case KeyEvent.VK_B:   //Synchronize vessel to driver.
               //Update of driver is instant based on coming V. update:
               //do2 using vFpAccIncr seems redundant: must be 0.
               HFold.oFp = Vec.fixAzi(HFold.vFpS + vFpAccIncr); 
               navigationMessage="DRIVER TURNED TO VESSEL";
               flagUpdatePosAScene = true;
               HFold.doPaintScratchMaps = true;
               break;

     
          //- - - - - - - - - - - - - - - - - - - - - 
          // Special observation switches.
          //------------------------------------------


          //------------------------------------------
          //User dialog keys:
          //- - - - - - - - - - - - - - - - - - - - - 
                     case KeyEvent.VK_Q:
                          Scenario.setText(Scenario.TXT_QUICK_HELP);
                          Stat.helpCalls++;
                          break;
                     case KeyEvent.VK_H:
                          Scenario.setText(Scenario.TXT_HELP);
                          Stat.helpCalls++;
                          break;
                     case KeyEvent.VK_I:
                          Scenario.setText(Scenario.TXT_INTRODUCTION);
                          break;
                     case KeyEvent.VK_1:
                          Scenario.setText(Scenario.TXT_LEVEL1HELP);
                          break;
                   
           //- - - - - - - - - - - - - - - - - - - - - 
           //User dialog keys:
           //------------------------------------------


           case KeyEvent.VK_E:
                displayAxes = !displayAxes;
                break;

           case KeyEvent.VK_X:
                //Terminate Application graciously:
                exitFlag = exitORDERED;
                exitLastFrame = frameNumber + exitDURATION;
                GS.con("Exit sequence initiated ... to frame = " + exitLastFrame);
                Scenario.setupGrandMessage("Thank you for Adventure ...", 
                                           frameNumber, 40, 1);
                break;





                     //==============================
                     //  DISABLED KEYS
                     //------------------------------
                     /*
                     //Not used. Disable:
                     case KeyEvent.VK_...:
                          Vessel.cp.setVisible(!Vessel.cp.isVisible());
                          break;
                     case KeyEvent.VK_T:
                          Vessel.cp.setVisible(!Vessel.cp.isVisible());
                          Vessel.consoleToTextArea = 
                          !Vessel.consoleToTextArea;
                          break;
                     */

                     


                     /*   Thisi a great way to make talking all of the pads.
                          This case toggles focus on pads.
                          However, we don't have use of this now, so 
                          disable this feature:
                     */
                     /*
                     //Not used. Disable>
                     case KeyEvent.VK_W:
                          //do2 This code should be static in SPad:
                          if( SPad.focusIx != SPad.none_ ) {
                              SPad.sPad[SPad.focusIx].removeFocus();
                          }
                          SPad.focusIx = ((SPad.focusIx+2)%(SPad.d_+1))-1; 
                          if( SPad.focusIx != SPad.none_ ) {
                              SPad.sPad[SPad.focusIx].setFocus();
                          }
                          break;
                     */
                     //------------------------------
                     //  DISABLED KEYS
                     //==============================



          }          //switch (keyCode) at Ctrl.
          break;     // case KeyEvent.CTRL_DOWN_MASK:

     default:  //not CTRL_DOWN_MASK
               if( keyCode == KeyEvent.VK_ENTER ) {
                   eventToScenario = Scenario.EV_MESSAGEFROMDRIVER;
                   return;
               } 

     } // switch (keyModifiers)


      //=================================================
      //  Backward and forward motion keys.
      //  Calculation and updating back/forward motions.
      //  Handling collisions and reversing back.
      //-------------------------------------------------
      if( keyModifiers != KeyEvent.CTRL_DOWN_MASK ) {
      if( keyCode == KeyEvent.VK_DOWN |
          keyCode == KeyEvent.VK_UP ) {

          double signedStep = stepXY;
          if( keyCode == KeyEvent.VK_DOWN ) {
              navigationMessage="BACK   ";
              signedStep = -stepXY;
              //oAccomulatedPath -= stepXY;
          } else {
              navigationMessage="FORWARD";
              //oAccomulatedPath += stepXY;
          }

          doTriesToAvoidCollisions(signedStep, true);

          /*
          //GS.con("Upd PosAScene bef. Coll. because of BACK/FORWARD keys");
          //updatePosAScene();
          z = x3;              
          if( CollisionMessage(z, 2 * elevationStep) < 0 ) {
                  while( CollisionMessage(z, 2 * elevationStep) < 0  && 
                         z < ceiling ) {
                         z = z + defaultStep;
                  }
                  if( z < ceiling ) {
                      x3 = z; //No move, just raising.
                      keyUpDownCode=KEY_UP_DOWN_DIGESTED;
                      //GS.con("Upd PosAScene because elevation by collision.");
                      updatePosAScene();
                  } else {
                    oAccomulatedPath = savedPath;   //Restore hf:
                    keyUpDownCode=KEY_UP_DOWN_COLLISION;
                    //GS.con("Upd PosAScene because restoring bec. of collis.");
                    updatePosAScene();
                  }
          } else {
                   keyUpDownCode=KEY_UP_DOWN_DIGESTED;
          }
          */


      }
      } //if keyModifiers != KeyEvent.CTRL_DOWN_MASK
      //-------------------------------------------------
      //  Backward and forward motion keys.
      //=================================================



      //=================================================
      //  Turns
      //-------------------------------------------------
      //
      //This can be an improvement:
      //Precautioner against lost stop-key:
      //However, hf-motion-fsm already has precautioner
      //for lost stop-keys:
      if( Math.abs(oAccomulatedPath) < 1.e-22 ) {
         
        if( keyCode == KeyEvent.VK_RIGHT ||
             keyCode == KeyEvent.VK_LEFT ){

             double loSignedTurn = AngleStep;
             navigationMessage=".LEFT   ";
             if( keyCode == KeyEvent.VK_RIGHT ) { 
                 loSignedTurn = -loSignedTurn;
                 navigationMessage=".RIGHT   ";
             }
             if( keyModifiers == KeyEvent.CTRL_DOWN_MASK ) {
                 navigationMessage = "DRIVER" + navigationMessage; 
                 HFold.oFp=Vec.fixAzi(HFold.oFp+loSignedTurn);
             }else{  
                 flagUpdateCurrentVesselPosAOrientation = true;
                 //do2 make less accuracy loss:
                 //double wStashDifference = HFold.oFp - vFpAccIncr;
                 vFpAccIncr = Vec.fixAzi(vFpAccIncr+loSignedTurn);
                 if(!autoPilot) { 
                    //HFold.oFp=Vec.fixAzi(wStashDifference+vFpAccIncr);
                    HFold.oFp=Vec.fixAzi(HFold.oFp+loSignedTurn);
                    navigationMessage = "BOTH" + navigationMessage; 
                 }else{
                    navigationMessage = "VESSEL" + navigationMessage; 
                 }
             }
             flagUpdateScenePopulation = true;
             HFold.doPaintScratchMaps = true;
             return;
         }

         // 90 d e g r e e    t u r n s:

         //do1b More precise method must be:
         //i = KeyEvent.SHIFT_DOWN_MASK | KeyEvent.CTRL_DOWN_MASK;
         //if( keyModifiers & i == i ) {
         if( keyModifiers == KeyEvent.CTRL_DOWN_MASK ) {
         if( keyCode == KeyEvent.VK_PERIOD ||
             keyCode == KeyEvent.VK_COMMA ){
            
             double loSignedTurn = Vec.PI2;
             navigationMessage=".RIGHT 90^   ";
             if( keyCode == KeyEvent.VK_PERIOD ) { 
                 loSignedTurn = -loSignedTurn;
                 navigationMessage=".RIGHT 90^  ";
             }
             //GS.con("SignedTurn=" + loSignedTurn + "vIncr=" + vFpAccIncr);
             vFpAccIncr = Vec.fixAzi(vFpAccIncr+loSignedTurn);
             //GS.con("      After. vIncr=" + vFpAccIncr);
             if(!autoPilot) { 
                HFold.oFp=Vec.fixAzi(HFold.oFp+loSignedTurn);
                navigationMessage = "BOTH" + navigationMessage; 
             }else{
                navigationMessage = "VESSEL" + navigationMessage; 
             }
             flagUpdateCurrentVesselPosAOrientation = true;
             flagUpdateScenePopulation = true;
             HFold.doPaintScratchMaps = true;
             return;
         }
         } // if( keyModifiers == KeyEvent.CTRL_DOWN_MASK

      } //if( Math.abs(oAccomul...


      //-------------------------------------------------
      // Altitutde, observational turn.
      //- - - - - - - - - - - - - - - - - - - - - - - - -
      if( engAltitude ) {
      if( keyModifiers == KeyEvent.CTRL_DOWN_MASK ) {
      if( keyCode == KeyEvent.VK_UP ||
          keyCode == KeyEvent.VK_DOWN ){

          double loSignedTurn = AngleStep;
          navigationMessage="ALTI.UP   ";
          if( keyCode == KeyEvent.VK_DOWN  ) { 
              loSignedTurn = -loSignedTurn;
              navigationMessage="ALTI.DOWN   ";
          }
          w = dAltiIncr + loSignedTurn;
          //Restrict Altitide by some arbitrary angle .8*Vec.PI2.
          if( Math.abs(w) < .8*Vec.PI2 ) dAltiIncr=w;
          return;
      }
      }
      }
      //- - - - - - - - - - - - - - - - - - - - - - - - -
      // Altitutde, observational turn.
      //--------------------------------------------------

      //-------------------------------------------------
      //  Turns
      //=================================================



        if( keyCode == KeyEvent.VK_PAGE_DOWN |
            keyCode == KeyEvent.VK_PAGE_UP     ) {

             if( keyCode == KeyEvent.VK_PAGE_DOWN ) {
                 st = -elevationStep;
                 navigationMessage="DOWN   ";  
             } else {
                 st = elevationStep;
                 navigationMessage="ELEVATE ";  
             }  
             z  = x3 + st;
             if( CollisionMessage(z,2*elevationStep ) <= 0 ) {
                 encourageObserver = false;
             } else {
               //protection:
               if( z <= ceiling && z > floor ) {
                   x3 = z;
                   //messys(s);
               } else {
                 encourageObserver = false;
                 if( z <= ceiling ) {
                     if( Scenario.level < Scenario.levHFConqueror ) {
                       mesc("Cannot travel undeground in this level."); 
                     } else {
                       x3 = z;
                       //messys(s);
                     }
                 } else {
                   if( Scenario.level < Scenario.levHFConqueror ) {
                       mesc("Your Vessel tried to ping the Elevation Sphere ... "); 
                       mes ("This behavior is forgivable for Humans ... ");
                       mes ("However, Ightarians consider this inpolite in resect to Gods ... "); 
                   } else {
                       x3 = z;
                       //messys(s);
                   }
                } 
              } //protection end
          } // if( CollisionMessage
        }

        if( keyModifiers != KeyEvent.CTRL_DOWN_MASK &&
            keyModifiers != KeyEvent.ALT_DOWN_MASK  ) {
            if( keyCode == KeyEvent.VK_ESCAPE ) {
                Scenario.closeSketchAndHelp();
                return; 
            }
        }

   }


//Auxiliary shortcut-methods:
public void mes(String s) {
       SPad.sPad[SPad.N_].wr(s);
}
public void mesc(String s) {
       SPad.sPad[SPad.N_].wrc(s);
}



public void messys(String s) {
       SPad.sPad[spSys].wr( String.valueOf((System.currentTimeMillis()%100000)) + " " + s);
       //textmes(s);
}
public void messysc(String s) {
       SPad.sPad[spSys].wrc( String.valueOf((System.currentTimeMillis()%100000)) + " " + s);
}

public void meslev(String s) {
       SPad.sPad[spModes].wr(s);
}
public void meslevc(String s) {
       SPad.sPad[spModes].wrc(s);
}

public void meshelp(String s) {
       SPad.sPad[spSys].wr(s);
       //textmes(s);
}


public double CollisionMessage( double z, double gap) {
       double d;
       //d = z - Pop.hills[0].sceneHeight() - gap;
       d = z - Pop.grHill.sceneHeight() - gap;
       if( d < 0) {
           //GS.con("collision:" + d);
           mesc ("YOUR VESSEL HIT THE SURFACE AND BOUNCED UP ... ");
           mes  ("Drive carefully ... ");
      } else {
           // messys("Collision sur. dist. " + ((int)d));
      }
      return d;
}

private void updatePosAScene() {
        HFold.updatePathLastMotion();
        Pop.PopulateScene();

        //do1 inconsistent to put this here:
        HFold.runAngleminerAnDerivator(); 

        flagUpdatePosAScene = false;
}






/**  Checks for collision and tries to elevate vessel if any. 
     Returns true if movement succeessfully made. False if not. 
*/
private boolean doTriesToAvoidCollisions(double stepLength, boolean changeKey) {
         double savedPath  = oAccomulatedPath;
         oAccomulatedPath += stepLength;
         updatePosAScene();
         double z = x3;              
         if( CollisionMessage(z, 2 * elevationStep) < 0 ) {
                  while( CollisionMessage(z, 2 * elevationStep) < 0  && 
                         z < ceiling ) {
                         z = z + defaultStep;
                  }
                  if( z < ceiling ) {
                      x3 = z; //No move, just raising.
                      if(changeKey) keyUpDownCode=KEY_UP_DOWN_DIGESTED;
                      //GS.con("Upd PosAScene because elevation by collision.");
                      updatePosAScene();
                      return true;
                  } else {
                    oAccomulatedPath = savedPath;   //Restore hf:
                    if(changeKey) keyUpDownCode=KEY_UP_DOWN_COLLISION;
                    //GS.con("Upd PosAScene because restoring bec. of collis.");
                    updatePosAScene();
                    return false;
                  }
         } else {
                if(changeKey) keyUpDownCode=KEY_UP_DOWN_DIGESTED;
                return true;
         }
}









//Full message:
public void mesnavF(String s) {
       if( autoPilot ) {
           SPad.sPad[spNav].wrc("VESSEL-NAVIGATION   BY-AUTOPILOT  " +  HFold.stReadable());
       } else {
           SPad.sPad[spNav].wrc("VESSEL-NAVIGATION   BY-DRIVER     " +  HFold.stReadable());
       }
       if( s != null ) {
           SPad.sPad[spNav].wr("COMMAND " + keyCode + "." + keyID + " " + s );
       } 
       mesnav(" ");
}



public void mesnav(String s) {

       SPad spn = SPad.sPad[spNav];
       spn.wr(s);
       HFold.diplayVesselMotionDigits(); 
       spn.wr("SPEEDS       " + stepXY + " " + "steps/click" + 
              "  AZI " + Vec.rad2Degr(AngleStep) + " ^/click");
       spn.wr("SPEEDS EXACT " + stepXY + "  " + AngleStep);
       spn.wr("ELEVATION    " + ((int)x3));
       if( Math.abs(dAltiIncr) > .009 ) {
           spn.wr("ALTI         " + (Vec.RAD2DEG*dAltiIncr) + "^");
       }
}




} //Class End